package org.naokishibata.graphics3d;

import org.naokishibata.generic.*;
import org.naokishibata.math.*;

public class ViewMatrix extends Matrix {
    public ViewMatrix() {
	super(4,4);
	a[0] = a[5] = a[10] = a[15] = 1;
    }

    public ViewMatrix(Matrix m) {
	super(4,4);
	for(int i=0;i<4;i++) {
	    for(int j=0;j<4;j++) a[i*4+j] = m.v(i,j);
	}
    }

    public ViewMatrix(double[] a) {
	super(a, 4, 4);
    }

    public static ViewMatrix identity() {
	return new ViewMatrix();
    }

    public ViewMatrix times(ViewMatrix m) { return new ViewMatrix( super.times(m) ); }
    public ViewMatrix plus (ViewMatrix m) { return new ViewMatrix( super.plus(m)  ); }
    public ViewMatrix minus(ViewMatrix m) { return new ViewMatrix( super.minus(m) ); }
    public ViewMatrix times(double d)     { return new ViewMatrix( super.times(d) ); }
    public ViewMatrix transposition()     { return new ViewMatrix( super.transposition() ); }
    //public ViewMatrix inversion()         { return new ViewMatrix( super.inversion() ); }

    private static class Mat2x2 {
	final double a0, a1, a2, a3;

	Mat2x2(double x, double y, double z, double w) {
	    a0 = x; a1 = y; a2 = z; a3 = w;
	}

	Mat2x2 plus(Mat2x2 m) {
	    return new Mat2x2(a0 + m.a0, a1 + m.a1, a2 + m.a2, a3 + m.a3);
	}

	Mat2x2 minus(Mat2x2 m) {
	    return new Mat2x2(a0 - m.a0, a1 - m.a1, a2 - m.a2, a3 - m.a3);
	}

	Mat2x2 times(Mat2x2 m) {
	    return new Mat2x2(a0 * m.a0 + a1 * m.a2, a0 * m.a1 + a1 * m.a3,
			      a2 * m.a0 + a3 * m.a2, a2 * m.a1 + a3 * m.a3);
	}

	Mat2x2 times(double d) {
	    return new Mat2x2(a0 * d, a1 * d, a2 * d, a3 * d);
	}

	Mat2x2 inv() {
	    double d = 1.0 / (a0*a3 - a1*a2);
	    return new Mat2x2(a3 * d, -a1 * d, -a2 * d, a0 * d);
	}
    }

    private static ViewMatrix inv(ViewMatrix m) {
	Mat2x2 p = new Mat2x2(m.v(0,0), m.v(0,1), m.v(1,0), m.v(1,1));
	Mat2x2 q = new Mat2x2(m.v(0,2), m.v(0,3), m.v(1,2), m.v(1,3));
	Mat2x2 r = new Mat2x2(m.v(2,0), m.v(2,1), m.v(3,0), m.v(3,1));
	Mat2x2 s = new Mat2x2(m.v(2,2), m.v(2,3), m.v(3,2), m.v(3,3));
	Mat2x2 si = s.inv(), qsi = q.times(si);
	Mat2x2 sir = si.times(r), qsir = qsi.times(r);
	Mat2x2 pt = p.minus(qsir).inv();
	Mat2x2 qt = pt.times(qsi).times(-1);
	Mat2x2 rt = sir.times(pt);
	Mat2x2 st = si.plus(rt.times(qsi));

	double[] a = {
	     pt.a0,  pt.a1, qt.a0, qt.a1,
	     pt.a2,  pt.a3, qt.a2, qt.a3,
	    -rt.a0, -rt.a1, st.a0, st.a1,
	    -rt.a2, -rt.a3, st.a2, st.a3,
	};

	return new ViewMatrix(a);
    }

    //public ViewMatrix inversion() { return inv(this); }
    public ViewMatrix inversion()         { return new ViewMatrix( super.inversion() ); }

    public static ViewMatrix translate(Vector3D v) {
	ViewMatrix m = new ViewMatrix();
	m.a[0] = 1; m.a[1] = 0; m.a[2] = 0; m.a[3] = v.x();
	m.a[4] = 0; m.a[5] = 1; m.a[6] = 0; m.a[7] = v.y();
	m.a[8] = 0; m.a[9] = 0; m.a[10] = 1; m.a[11] = v.z();
	m.a[12] = 0; m.a[13] = 0; m.a[14] = 0; m.a[15] =     1;
	return m;
    }

    public static ViewMatrix scale(double d) {
	ViewMatrix m = new ViewMatrix();
	m.a[0] = d; m.a[1] = 0; m.a[2] = 0; m.a[3] = 0;
	m.a[4] = 0; m.a[5] = d; m.a[6] = 0; m.a[7] = 0;
	m.a[8] = 0; m.a[9] = 0; m.a[10] = d; m.a[11] = 0;
	m.a[12] = 0; m.a[13] = 0; m.a[14] = 0; m.a[15] = 1;
	return m;
    }

    public static ViewMatrix scale(double sx, double sy, double sz) {
	ViewMatrix m = new ViewMatrix();
	m.a[0] = sx; m.a[1] = 0; m.a[2] = 0; m.a[3] = 0;
	m.a[4] = 0; m.a[5] = sy; m.a[6] = 0; m.a[7] = 0;
	m.a[8] = 0; m.a[9] = 0; m.a[10] = sz; m.a[11] = 0;
	m.a[12] = 0; m.a[13] = 0; m.a[14] = 0; m.a[15] = 1;
	return m;
    }

    public static ViewMatrix rotX(double d) {
	ViewMatrix m = new ViewMatrix();
	double c = Math.cos(d), s = Math.sin(d);
	m.a[0] =  1; m.a[1] =  0; m.a[2] =  0; m.a[3] = 0;
	m.a[4] =  0; m.a[5] =  c; m.a[6] = -s; m.a[7] = 0;
	m.a[8] =  0; m.a[9] =  s; m.a[10] =  c; m.a[11] = 0;
	m.a[12] =  0; m.a[13] =  0; m.a[14] =  0; m.a[15] = 1;
	return m;
    }

    public static ViewMatrix rotY(double d) {
	ViewMatrix m = new ViewMatrix();
	double c = Math.cos(d), s = Math.sin(d);
	m.a[0] =  c; m.a[1] =  0; m.a[2] =  s; m.a[3] = 0;
	m.a[4] =  0; m.a[5] =  1; m.a[6] =  0; m.a[7] = 0;
	m.a[8] = -s; m.a[9] =  0; m.a[10] =  c; m.a[11] = 0;
	m.a[12] =  0; m.a[13] =  0; m.a[14] =  0; m.a[15] = 1;
	return m;
    }

    public static ViewMatrix rotZ(double d) {
	ViewMatrix m = new ViewMatrix();
	double c = Math.cos(d), s = Math.sin(d);
	m.a[0] =  c; m.a[1] = -s; m.a[2] =  0; m.a[3] = 0;
	m.a[4] =  s; m.a[5] =  c; m.a[6] =  0; m.a[7] = 0;
	m.a[8] =  0; m.a[9] =  0; m.a[10] =  1; m.a[11] = 0;
	m.a[12] =  0; m.a[13] =  0; m.a[14] =  0; m.a[15] = 1;
	return m;
    }

    public static ViewMatrix rotateBy(double angle, Vector3D center, Vector3D axisDir) {
        ViewMatrix mat;

        double a0 = Math.atan2(axisDir.y(), axisDir.x());
        Vector3D t = ViewMatrix.rotZ(-a0).times(axisDir);
        double a1 = -Math.atan2(t.z(), t.x());

        mat = ViewMatrix.rotY(-a1).times(ViewMatrix.rotZ(-a0));
        mat = ViewMatrix.rotX(angle).times(mat);
        mat = ViewMatrix.rotZ( a0).times(ViewMatrix.rotY( a1).times(mat));

	return mat;
    }

    public static ViewMatrix frustum(double l, double r, double b, double t, double n, double f) {
	ViewMatrix m = new ViewMatrix();
	m.a[0] = 2*n/(r-l); m.a[1] =           0; m.a[2] =  (r+l)/(r-l); m.a[3] = 0;
	m.a[4] =         0; m.a[5] = (2*n)/(t-b); m.a[6] =  (t+b)/(t-b); m.a[7] = 0;
	m.a[8] =         0; m.a[9] =           0; m.a[10] = -(f+n)/(f-n); m.a[11] = -2*f*n/(f-n);
	m.a[12] =         0; m.a[13] =           0; m.a[14] =           -1; m.a[15] = 0;
	return m;
    }

    public static ViewMatrix orthographic(double l, double r, double b, double t, double n, double f) {
	ViewMatrix m = new ViewMatrix();
	m.a[0 ] = 2/(r-l); m.a[1 ] =       0; m.a[2 ] =        0; m.a[3 ] = -(r+l)/(r-l);
	m.a[4 ] =       0; m.a[5 ] = 2/(t-b); m.a[6 ] =        0; m.a[7 ] = -(t+b)/(t-b);
	m.a[8 ] =       0; m.a[9 ] =       0; m.a[10] = -2/(f-n); m.a[11] = -(f+n)/(f-n);
	m.a[12] =       0; m.a[13] =       0; m.a[14] =        0; m.a[15] = 1;
	return m;
    }

    public Vector3D times(Vector3D v) {
	return new Vector3D(a[0]*v.x() + a[1]*v.y() + a[2]*v.z() + a[3],
			    a[4]*v.x() + a[5]*v.y() + a[6]*v.z() + a[7],
			    a[8]*v.x() + a[9]*v.y() + a[10]*v.z() + a[11]);
    }

    /*
    public Vector4 times(Vector4 v) {
	return new Vector4(a[0]*v.x() + a[1]*v.y() + a[2]*v.z() + a[3]*v.w(),
			   a[4]*v.x() + a[5]*v.y() + a[6]*v.z() + a[7]*v.w(),
			   a[8]*v.x() + a[9]*v.y() + a[10]*v.z() + a[11]*v.w(),
			   a[12]*v.x() + a[13]*v.y() + a[14]*v.z() + a[15]*v.w());
    }

    public static Vector4 perspectiveDivision(Vector4 v) {
        return new Vector4(v.x()/v.w(), v.y()/v.w(), v.z()/v.w(), 1);
    }

    public static Vector3D viewportTransform3(Vector4 v, ViewportSetting vs) {
        return new Vector3D(vs.x()+vs.width()/2+v.x()*vs.width()/2, vs.y()+vs.height()/2+v.y()*vs.height()/2, v.z());
    }

    public static Vector3D viewportTransform3IY(Vector4 v, ViewportSetting vs) {
        return new Vector3D(vs.x()+vs.width()/2+v.x()*vs.width()/2, vs.y()+vs.height()/2-v.y()*vs.height()/2, v.z());
    }

    public static Vector3D project3(Vector3D v, ViewMatrix mmat, ViewMatrix pmat, ViewportSetting vs) {
	return viewportTransform3(perspectiveDivision(pmat.times(mmat.times(new Vector4(v)))),vs);
    }

    public static Vector3D project3IY(Vector3D v, ViewMatrix mmat, ViewMatrix pmat, ViewportSetting vs) {
	return viewportTransform3IY(perspectiveDivision(pmat.times(mmat.times(new Vector4(v)))),vs);
    }
    */

    /*
    public static Vector2D viewportTransform(Vector4 v, ViewportSetting vs) {
        return new Vector2D(vs.x()+vs.width()/2+v.x()*vs.width()/2, vs.y()+vs.height()/2+v.y()*vs.height()/2);
    }

    public static Vector2D viewportTransformIY(Vector4 v, ViewportSetting vs) {
        return new Vector2D(vs.x()+vs.width()/2+v.x()*vs.width()/2, vs.y()+vs.height()/2-v.y()*vs.height()/2);
    }

    public static Vector2D project(Vector3D v, ViewMatrix mmat, ViewMatrix pmat, ViewportSetting vs) {
	return viewportTransform(perspectiveDivision(pmat.times(mmat.times(new Vector4(v)))),vs);
    }

    public static Vector2D projectIY(Vector3D v, ViewMatrix mmat, ViewMatrix pmat, ViewportSetting vs) {
	return viewportTransformIY(perspectiveDivision(pmat.times(mmat.times(new Vector4(v)))),vs);
    }
    */

    public static ViewMatrix lookAt(Vector3D eye, Vector3D center, Vector3D up, double sc) {
	Vector3D d = Vector3D.sub(center, eye);
	Vector3D u = up;
	Vector3D t;

	ViewMatrix ret;

	ret = rotZ(Math.PI/2 - Math.atan2(d.y(), d.x()));

	t = ret.times(d);

	ret = rotX(-Math.PI/2 - Math.atan2(t.z(), t.y())).times(ret);

	t = ret.times(u);

	ret = rotZ(Math.PI/2 - Math.atan2(t.y(), t.x())).times(ret);

	ret = ret.times(scale(sc));

	ret = translate(ret.times(eye).times(-1)).times(ret);

	return ret;
    }

    public static Vector3D projectNormalVector(ViewMatrix mat, Vector3D n) {
	Vector3D px = n.cross(new Vector3D(1,0,0));
	Vector3D py = n.cross(new Vector3D(0,1,0));
	Vector3D pz = n.cross(new Vector3D(0,0,1));

	double lx = px.lengthSquared(), ly = py.lengthSquared(), lz = pz.lengthSquared();

	Vector3D p;

	if (lx >= ly && lx >= lz) {
	    p = px;
	} else if (ly >= lz) {
	    p = py;
	} else {
	    p = pz;
	}

	Vector3D q = n.cross(p);
	Vector3D r = n.cross(q);

	Vector3D t0 = mat.times(new Vector3D(0,0,0));
	Vector3D t1 = mat.times(q);
	Vector3D t2 = mat.times(r);

	return (t1.minus(t0)).cross(t2.minus(t0)).normalized();
    }

    public double[] mat16() {
	double ret[] = new double[16];

	for(int i=0;i<4;i++) {
	    for(int j=0;j<4;j++) ret[i+j*4] = a[i*4+j];
	}

	return ret;
    }
}
