package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

public class Line3D implements Serializable {
    private final Vector3D u[];

    private void check() {
	assert u[0] != null && u[1] != null;
	assert !Vector3D.epsilonEquals(u[0],u[1]);
    }

    public Line3D(Vector3D u0, Vector3D u1) {
	u    = new Vector3D[2];
	u[0] = u0;
	u[1] = u1;
    }

    public Line3D(LineSegment3D l) {
	u    = new Vector3D[2];
	u[0] = l.v(0);
	u[1] = l.v(1);
    }

    public String toString() {
	return "(Line "+u[0].toString()+"-"+u[1].toString()+")";
    }

    public Vector3D v(int i) {
	return u[i];
    }

    //

    public static boolean parallel(Line3D l, Line3D m, double eps) {
	Vector3D v0 = Vector3D.normalize(Vector3D.sub(l.u[0],l.u[1]));
	Vector3D v1 = Vector3D.normalize(Vector3D.sub(m.u[0],m.u[1]));

	return Math.abs(Vector3D.crossProduct(v0,v1).length()) < eps;
    }

    public static boolean parallel(Line3D l, Line3D m) {
	return parallel(l, m, MathUtils.angleEpsilon);
    }

    public static boolean parallel(Line3D l, Plane3D m, double eps) {
	return Plane3D.parallel(m,l,eps);
    }

    public static double distance(Line3D l, Vector3D v) {
	double d = Vector3D.length(Vector3D.sub(l.u[1],l.u[0]));
	if (d == 0) return Vector3D.distance(l.u[0],v);
	double n = Vector3D.length(Vector3D.crossProduct(Vector3D.sub(l.u[1],l.u[0]),Vector3D.sub(l.u[0],v)));

	return n / d;
    }

    public static boolean includes(Line3D l, Vector3D v) {
	return distance(l,v) <= MathUtils.epsilon;
    }

    public static double distance(Line3D l, Line3D m) {
	if (parallel(l,m)) return distance(m,l.u[0]);
	Vector3D a = Vector3D.sub(m.u[1],m.u[0]);
	Vector3D b = Vector3D.sub(l.u[1],l.u[0]);
	Vector3D c = Vector3D.sub(l.u[0],m.u[0]);
	double n = Math.abs(Vector3D.dotProduct(Vector3D.crossProduct(a,b),c));
	double d = Vector3D.length(Vector3D.crossProduct(a,b));

	return n / d;
    }

    /*
    public static Vector3D nearestPoint(Line3D l, Vector3D v) {
	double p = Vector3D.dotProduct(Vector3D.sub(l.u[1],l.u[0]),Vector3D.sub(v,l.u[0])) / Vector3D.distanceSquared(l.u[1],l.u[0]);
	return Vector3D.add(Vector3D.scale(p,Vector3D.sub(l.u[1],l.u[0])),l.u[0]);
    }

    public double distanceFrom(LineSegment3D l) {
	return LineSegment3D.distance(l,this);
    }

    public static LineSegment3D nearestPointsBetween(Line3D m, Line3D l) {
	Vector3D u00 = m.u[0];
	Vector3D u01 = Vector3D.sub(m.u[1],m.u[0]);
	Vector3D u10 = l.u[0];
	Vector3D u11 = Vector3D.sub(l.u[1],l.u[0]);

	//if (u01.cross(l.u[0].minus(m.u[0])).dot(l.u[1].minus(m.u[0])) < MathUtils.epsilon) {
	if (Vector3D.dotProduct(Vector3D.crossProduct(u01, Vector3D.sub(l.u[0],m.u[0])),Vector3D.sub(l.u[1],m.u[0])) < MathUtils.epsilon) {
	    // four points are on a plane
	    //if (parallel(m,l)) return new LineSegment3D(m.u[0],nearestPoint(l, m.u[0]));

	    Vector3D lv = distance(m, l.u[0]) > distance(m, l.u[1]) ? l.u[0] : l.u[1];

	    Plane3D p = new Plane3D(m.u[0],m.u[1],
				    Vector3D.add(Vector3D.crossProduct(Vector3D.sub(m.u[1],m.u[0]),
								       Vector3D.sub(lv    ,m.u[0])),
						 m.u[0]));

	    Vector3D v1 = Plane3D.intersection(p,l);

	    if (v1 == null) {
		// four points are on a plane
		return new LineSegment3D(m.u[0],nearestPoint(l, m.u[0]));
	    }

	    Vector3D v0 = nearestPoint(m,v1);

	    return new LineSegment3D(v0,v1);
	}

	double a[][] = new double[2][];
	a[0] = new double[3];
	a[1] = new double[3];

	a[0][0] = Vector3D.dotProduct(u01,u01);
	a[0][1] = Vector3D.dotProduct(u01,u11);
	a[0][2] = Vector3D.dotProduct(u01,Vector3D.sub(u10,u00));
	a[1][0] = Vector3D.dotProduct(u11,u01);
	a[1][1] = Vector3D.dotProduct(u11,u11);
	a[1][2] = Vector3D.dotProduct(u11,Vector3D.sub(u10,u00));

	GaussJordanMethod.solve(a);

	Vector3D v0 = Vector3D.add(u00,Vector3D.scale(a[0][2],u01));
	Vector3D v1 = Vector3D.add(u10,Vector3D.scale(a[1][2],u11));

	return new LineSegment3D(v0,v1);
    }

    public Vector3D nearestPoint(Line3D l) {
	return nearestPointsBetween(this,l).v(0);
    }

    public Vector3D nearestPointFrom(Vector3D v) {
	return nearestPoint(this, v);
    }
    */

    //

    public boolean isParallelTo(Line3D l, double eps) {
	return parallel(this, l, eps);
    }

    public boolean isParallelTo(Line3D l) {
	return parallel(this, l, MathUtils.angleEpsilon);
    }

    public boolean isParallelTo(Plane3D p, double eps) {
	return Plane3D.parallel(p,this,eps);
    }

    public boolean isParallelTo(Plane3D p) {
	return Plane3D.parallel(p, this, MathUtils.angleEpsilon);
    }

    public double distanceFrom(Vector3D v) {
	return distance(this, v);
    }

    public double distanceFrom(Line3D l) {
	return distance(this, l);
    }
}
