package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

public class Plane3D implements Serializable {
    //private static final long serialVersionUID = MiscUtils.svuid();

    private final Vector3D u[];
    private final Vector3D normal;

    public Plane3D(Vector3D v0, Vector3D v1, Vector3D v2) {
	u    = new Vector3D[3];
	u[0] = v0;
	u[1] = v1;
	u[2] = v2;
	normal = Vector3D.normalize(Vector3D.crossProduct(Vector3D.sub(u[1],u[0]),Vector3D.sub(u[2],u[0])));
    }

    /*
    public Plane3D(Triangle3D t) {
	u    = new Vector3D[3];
	u[0] = t.v(0);
	u[1] = t.v(1);
	u[2] = t.v(2);
	normal = Vector3D.normalize(Vector3D.crossProduct(Vector3D.sub(u[1],u[0]),Vector3D.sub(u[2],u[0])));
    }
    */

    public Vector3D v(int i) {
	return u[i];
    }

    public String toString() {
	return "(Plane "+v(0).toString()+","+v(1).toString()+","+v(2).toString()+")";
    }

    //

    public static Vector3D normalVector(Plane3D p) { return p.normal; }

    public static Vector3D intersection(Plane3D p, Line3D l) {
	Vector3D n = Plane3D.normalVector(p);
	Vector3D v = Vector3D.normalize(Vector3D.sub(l.v(1), l.v(0)));
	double c = Vector3D.dotProduct(n,v);

	if (MathUtils.epsilonEquals(c,0,MathUtils.angleEpsilon)) {
	    if (Plane3D.includes(p,l.v(0))) return l.v(0);
	    return null;
	}

	double r = -(Vector3D.dotProduct(n,l.v(0)) - Vector3D.dotProduct(n,p.v(0))) / c;
	return Vector3D.add(l.v(0), Vector3D.scale(v, r));
    }

    public static Vector3D intersection(Line3D l, Plane3D p) {
	return intersection(p,l);
    }

    public static Line3D intersection(Plane3D q, Plane3D p) {
	Vector3D n = normalVector(q);
	Vector3D a = Vector3D.crossProduct(n,normalVector(p));
	Vector3D b = Vector3D.crossProduct(n,a);

	Line3D l = new Line3D(q.u[0],Vector3D.add(q.u[0],b));
	Vector3D x = intersection(p,l);

	return new Line3D(x, Vector3D.add(x,a));
    }

    public static Vector3D nearestPoint(Plane3D p, Vector3D v) {
	return intersection(p, new Line3D(v,Vector3D.add(v,normalVector(p))));
    }

    public static double distance(Plane3D p, Vector3D v) {
	Vector3D n = Vector3D.crossProduct(Vector3D.sub(p.u[1],p.u[0]),Vector3D.sub(p.u[2],p.u[0]));
	double d = -Vector3D.dotProduct(n,p.u[0]);
	
	return Math.abs(Vector3D.dotProduct(n,v) + d) / n.length();
    }

    public static int intersects(Plane3D p, LineSegment3D ls) {
	Vector3D v = intersection(p, new Line3D(ls));
	if (v == null) return 0;
	return LineSegment3D.includes(ls, v);
    }

    public static boolean includes(Plane3D p, Vector3D v, double eps) {
	if (p.v(0).epsilonEquals(v,eps) || p.v(1).epsilonEquals(v,eps) || p.v(2).epsilonEquals(v,eps)) return true;
	return distance(p,v) < eps;
    }

    public static boolean includes(Plane3D p, Vector3D v) {
	return includes(p,v,MathUtils.epsilon);
    }

    public static boolean includes(Plane3D p, LineSegment3D ls) {
	return includes(p,ls.v(0)) && includes(p,ls.v(1));
    }

    public static boolean parallel(Plane3D p, Line3D l, double eps) {
	Vector3D v0 = Vector3D.normalize(Vector3D.sub(l.v(0),l.v(1)));
	Vector3D v1 = normalVector(p);

	return Math.abs(Vector3D.dotProduct(v0, v1)) < eps;
    }

    public static boolean parallel(Plane3D p, Line3D l) {
	return parallel(p, l, MathUtils.angleEpsilon);
    }

    public static boolean parallel(Plane3D q, Plane3D p, double eps) {
	return Math.abs(Vector3D.length(Vector3D.crossProduct(normalVector(p),normalVector(q)))) < eps;
    }

    public static boolean parallel(Plane3D q, Plane3D p) {
	return parallel(q, p, MathUtils.angleEpsilon);
    }

    public static boolean definable(Vector3D v0, Vector3D v1, Vector3D v2) {
	double d0 = Vector3D.distance(v1,v2);
	double d1 = Vector3D.distance(v2,v0);
	double d2 = Vector3D.distance(v0,v1);

	if (d0 >= d1 && d0 >= d2) {
	    if (d0 <= MathUtils.epsilon) return false;
	    return !Line3D.includes(new Line3D(v1,v2), v0);
	} else if (d1 >= d0 && d1 >= d2) {
	    if (d1 <= MathUtils.epsilon) return false;
	    return !Line3D.includes(new Line3D(v2,v0), v1);
	} else {
	    if (d2 <= MathUtils.epsilon) return false;
	    return !Line3D.includes(new Line3D(v0,v1), v2);
	}
    }

    //

    public Vector3D normalVector() {
	return normalVector(this);
    }

    public Vector3D intersectionWith(Line3D l) {
	return intersection(this,l);
    }

    public Line3D intersectionWith(Plane3D p) {
	return intersection(this, p);
    }

    public Vector3D nearestPointFrom(Vector3D v) {
	return nearestPoint(this, v);
    }

    public boolean includes(Vector3D v, double eps) {
	return includes(this, v, eps);
    }

    public boolean includes(Vector3D v) {
	return includes(this, v);
    }

    public double distanceFrom(Vector3D v) {
	return distance(this, v);
    }

    public int intersects(LineSegment3D ls) {
	return intersects(this, ls);
    }

    public boolean isParallelTo(Line3D l, double eps) {
	return parallel(this, l, eps);
    }

    public boolean isParallelTo(Line3D l) {
	return parallel(this, l, MathUtils.angleEpsilon);
    }

    public boolean isParallelTo(Plane3D p) {
	return parallel(this, p, MathUtils.angleEpsilon);
    }

    /*
    // ax + by + cz + d == 0
    public double[] getEquation() {
	double a, b, c, d;

	{
	    double[] ary = {
		1, u[0].y(), u[0].z(),
		1, u[1].y(), u[1].z(),
		1, u[2].y(), u[2].z(),
	    };

	    a = MathUtils.determinant3x3(ary);
	}

	{
	    double[] ary = {
		u[0].x(), 1, u[0].z(),
		u[1].x(), 1, u[1].z(),
		u[2].x(), 1, u[2].z(),
	    };

	    b = MathUtils.determinant3x3(ary);
	}

	{
	    double[] ary = {
		u[0].x(), u[0].y(), 1,
		u[1].x(), u[1].y(), 1,
		u[2].x(), u[2].y(), 1,
	    };

	    c = MathUtils.determinant3x3(ary);
	}

	{
	    double[] ary = {
		u[0].x(), u[0].y(), u[0].z(),
		u[1].x(), u[1].y(), u[1].z(),
		u[2].x(), u[2].y(), u[2].z(),
	    };

	    d = -MathUtils.determinant3x3(ary);
	}

	double[] ret = {a, b, c, d};

	return ret;
    }
    */
}
