package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

/**
 * Vector2D is an immutable class representing a 2-dimensional vector.
 */
public class Vector2D implements Serializable, Comparable {
    /**
       The first coordinate.
     */
    public final double x;

    /**
       The first coordinate.
     */
    public final double y;

    /**
       Constructs a new Vector2D instance which represents (x0, y0) vector.
     */
    public Vector2D(double x0, double y0) {
	x = x0; y = y0;
    }

    /**
       Returns the first coordinate.
     */
    public double x() {
	return x;
    }

    /**
       Returns the second coordinate.
     */
    public double y() {
	return y;
    }

    /**
       Returns the string representation of the vector.
     */
    public String toString() {
	return new String("(" + x + "," + y + ")");
    }

    /**
       Returns the addition of vectors v and u.
     */
    public static Vector2D add(Vector2D v, Vector2D u) {
	return new Vector2D(v.x + u.x, v.y + u.y);
    }

    /**
       Returns the addition of vectors v, u and w.
     */
    public static Vector2D add(Vector2D v, Vector2D u, Vector2D w) {
	return new Vector2D(v.x + u.x + w.x, v.y + u.y + w.y);
    }

    /**
       Returns the addition of vectors v, u, w and x.
     */
    public static Vector2D add(Vector2D v, Vector2D u, Vector2D w, Vector2D x) {
	return new Vector2D(v.x + u.x + w.x + x.x, v.y + u.y + w.y + x.y);
    }

    /**
       Returns v - u.
     */
    public static Vector2D sub(Vector2D v, Vector2D u) {
	return new Vector2D(v.x - u.x, v.y - u.y);
    }

    /**
       Returns the dot product of vectors v and u.
     */
    public static double dotProduct(Vector2D v, Vector2D u) {
	return v.x*u.x + v.y*u.y;
    }

    /**
       Returns the cross product of vectors v and u.
     */
    public static double crossProduct(Vector2D v, Vector2D u) {
	return v.x*u.y - v.y*u.x;
    }

    /**
       Returns vector v scaled by s
     */
    public static Vector2D scale(double s, Vector2D v) {
	return new Vector2D(s * v.x, s * v.y);
    }

    /**
       Returns vector v scaled by s
     */
    public static Vector2D scale(Vector2D v, double s) {
	return new Vector2D(s * v.x, s * v.y);
    }

    /**
       Returns the normalized vector of v.
     */
    public static Vector2D normalize(Vector2D v) {
	double l = v.length();
	if (l != 0) return new Vector2D(v.x/l, v.y/l);

	return new Vector2D(1, 0);
    }

    /**
       Returns the length of v.
     */
    public static double length(Vector2D v) {
	return Math.sqrt(v.x*v.x + v.y*v.y);
    }

    /**
       Returns the squared length of v.
     */
    public static double lengthSquared(Vector2D v) {
	return v.x*v.x + v.y*v.y;
    }

    /**
       Returns the distance between v and w.
     */
    public static double distance(Vector2D v, Vector2D w) {
	return length(sub(v,w));
    }

    /**
       Returns the squared distance between v and w.
     */
    public static double distanceSquared(Vector2D v, Vector2D w) {
	return lengthSquared(sub(v,w));
    }

    /**
       Returns the midpoint between v and w.
     */
    public static Vector2D midpoint(Vector2D v, Vector2D w) {
	return scale(0.5, add(v,w));
    }

    /**
       Returns the angle between v and w.
     */
    public static double angle(Vector2D v, Vector2D w) {
	double a = Math.atan2(w.y, w.x) - Math.atan2(v.y, v.x);
	if (a >  Math.PI) return a - Math.PI*2;
	if (a < -Math.PI) return a + Math.PI*2;
	return a;
    }

    public static boolean epsilonEquals(Vector2D v, Vector2D u, double eps) {
	return distance(u,v) <= eps;
    }

    public static boolean epsilonEquals(Vector2D v, Vector2D u) {
	return distance(u, v) <= MathUtils.epsilon;
    }

    public static boolean equals(Vector2D v, Vector2D u) {
	return v.x == u.x && v.y == u.y;
    }

    //

    /**
       Returns the length of this vector.
     */
    public double length() {
	return Math.sqrt(x*x + y*y);
    }

    public Vector2D normalize() {
	double l = length();
	if (l != 0) return new Vector2D(x/l, y/l);

	return new Vector2D(1, 0);
    }

    public boolean epsilonEquals(Vector2D v, double eps) {
	return distance(this,v) <= eps;
    }

    public boolean epsilonEquals(Vector2D v) {
	return distance(this,v) <= MathUtils.epsilon;
    }

    /**
       Returns this + v
     */
    public Vector2D plus(Vector2D v) {
	return add(this,v);
    }

    /**
       Returns this - v
     */
    public Vector2D minus(Vector2D v) {
	return sub(this,v);
    }

    /**
       Returns the cross product of this and v
     */
    public double dot(Vector2D v) {
	return dotProduct(this,v);
    }

    public Vector2D dot(double d) {
	return scale(d,this);
    }

    public double cross(Vector2D v) {
	return crossProduct(this,v);
    }

    public Vector2D orthogonal() {
	return new Vector2D(y,-x);
    }

    public Vector2D normalized() {
	return normalize(this);
    }
    //

    public int hashCode() {
	return (int)(Double.doubleToRawLongBits(x) ^ Double.doubleToRawLongBits(y));
    }

    public int compareTo(Object o) {
	Vector2D v = (Vector2D) o;

	if (x > v.x) {
	    return 1;
	} else if (x < v.x) {
	    return -1;
	} else if (y > v.y) {
	    return 1;
	} else if (y < v.y) {
	    return -1;
	}

	return 0;
    }

    public boolean equals(Vector2D v) {
	return v.x == x && v.y == y;
    }

    public boolean equals(Object o) {
	return equals((Vector2D) o);
    }
}
