package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

/**
 * Vector3D is an immutable class representing a 3-dimensional vector.
 */
public class Vector3D implements Serializable, Comparable {
    /**
       The first coordinate.
     */
    public final double x;

    /**
       The second coordinate.
     */
    public final double y;

    /**
       The third coordinate.
     */
    public final double z;

    /**
       Constructs a new Vector3D instance which represents (x0, y0, z0) vector.
     */
    public Vector3D(double x0, double y0, double z0) {
	x = x0; y = y0; z = z0;
    }

    /**
       Returns the first coordinate of the vector.
     */
    public double x() {
	return x;
    }

    /**
       Returns the second coordinate of the vector.
     */
    public double y() {
	return y;
    }

    /**
       Returns the third coordinate of the vector.
     */
    public double z() {
	return z;
    }

    /**
       Returns the string representation of the vector.
     */
    public String toString() {
	return new String("(" + x + "," + y + "," + z + ")");
    }
    
    //

    /**
       Returns the addition of vectors v and u.
     */
    public static Vector3D add(Vector3D v, Vector3D u) {
	return new Vector3D(v.x + u.x, v.y + u.y, v.z + u.z);
    }

    /**
       Returns the addition of vectors v, u and w.
     */
    public static Vector3D add(Vector3D v, Vector3D u, Vector3D w) {
	return new Vector3D(v.x+u.x+w.x, v.y+u.y+w.y, v.z+u.z+w.z);
    }

    /**
       Returns the addition of vectors v, u, w and x.
     */
    public static Vector3D add(Vector3D v,Vector3D u,Vector3D w,Vector3D x) {
	return new Vector3D(v.x+u.x+w.x+x.x, v.y+u.y+w.y+x.y, v.z+u.z+w.z+x.z);
    }

    /**
       Returns vector v subtracted u.
     */
    public static Vector3D sub(Vector3D v, Vector3D u) {
	return new Vector3D(v.x - u.x, v.y - u.y, v.z - u.z);
    }

    /**
       Returns the dot product of vectors v and u.
     */
    public static double dotProduct(Vector3D v, Vector3D u) {
	return v.x*u.x + v.y*u.y + v.z*u.z;
    }

    /**
       Returns the cross product of vectors v and u.
     */
    public static Vector3D crossProduct(Vector3D v1, Vector3D v2) {
	return new Vector3D(v1.y * v2.z - v1.z * v2.y, v1.z * v2.x - v1.x * v2.z, v1.x * v2.y - v1.y * v2.x);
    }

    /**
       Returns vector v scaled by s
     */
    public static Vector3D scale(double s, Vector3D v) {
	return new Vector3D(s * v.x, s * v.y, s * v.z);
    }

    /**
       Returns vector v scaled by s
     */
    public static Vector3D scale(Vector3D v, double s) {
	return scale(s,v);
    }

    /**
       Returns the normalized vector of v.
     */
    public static Vector3D normalize(Vector3D v) {
	double l = length(v);
	if (l != 0) return new Vector3D(v.x/l, v.y/l, v.z/l);

	return new Vector3D(1, 0, 0);
    }

    /**
       Returns the length of vector v
     */
    public static double length(Vector3D v) {
	return Math.sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
    }

    /**
       Returns the squared length of vector v
     */
    public static double lengthSquared(Vector3D v) {
	return v.x*v.x + v.y*v.y + v.z*v.z;
    }

    /**
       Returns the distance between v and w
     */
    public static double distance(Vector3D v, Vector3D w) {
	return length(sub(v,w));
    }

    /**
       Returns the squared distance between v and w
     */
    public static double distanceSquared(Vector3D v, Vector3D w) {
	return lengthSquared(sub(v,w));
    }

    /**
       Returns the midpoint between v and w
     */
    public static Vector3D midpoint(Vector3D v, Vector3D w) {
	return scale(0.5, add(v,w));
    }

    /**
       Returns the centroid of v and w
     */
    public static Vector3D centroid(Vector3D v, Vector3D w) {
	return scale(0.5, add(v,w));
    }

    /**
       Returns the centroid of v, w and x
     */
    public static Vector3D centroid(Vector3D v, Vector3D w, Vector3D x) {
	return scale(1.0/3, add(v,w,x));
    }

    /**
       Returns the centroid of v, w, x and y
     */
    public static Vector3D centroid(Vector3D v, Vector3D w, Vector3D x, Vector3D y) {
	return scale(1.0/4, add(v,w,x,y));
    }

    /**
       Returns the angle of vector v and w
     */
    public static double angle(Vector3D v, Vector3D w) {
	return Math.acos(dotProduct(normalize(v),normalize(w)));
    }

    /**
       Returns vector v rotated on x-axis around (bx, by, bz)
     */
    public static Vector3D rotX(Vector3D v, double rx, double bx, double by, double bz) {
	double c = Math.cos(rx), s = Math.sin(rx);
	double y0 = v.y - by;
	double z0 = v.z - bz;
	return new Vector3D(v.x, by + y0*c - z0*s, bz + y0*s + z0*c);
    }

    /**
       Returns vector v rotated on y-axis around (bx, by, bz)
     */
    public static Vector3D rotY(Vector3D v, double ry, double bx, double by, double bz) {
	double c = Math.cos(ry), s = Math.sin(ry);
	double z0 = v.z - bz;
	double x0 = v.x - bx;
	return new Vector3D(bx + z0*s + x0*c, v.y, bz + z0*c - x0*s);
    }

    /**
       Returns vector v rotated on z-axis around (bx, by, bz)
     */
    public static Vector3D rotZ(Vector3D v, double rz, double bx, double by, double bz) {
	double c = Math.cos(rz), s = Math.sin(rz);
	double x0 = v.x - bx;
	double y0 = v.y - by;
	return new Vector3D(bx + x0*c - y0*s, by + x0*s + y0*c, v.z);
    }

    /**
       Returns vector v rotated on x-axis around b
     */
    public static Vector3D rotX(Vector3D v, double r, Vector3D b) { return rotX(v, r, b.x(), b.y(), b.z()); }

    /**
       Returns vector v rotated on y-axis around b
     */
    public static Vector3D rotY(Vector3D v, double r, Vector3D b) { return rotY(v, r, b.x(), b.y(), b.z()); }

    /**
       Returns vector v rotated on z-axis around b
     */
    public static Vector3D rotZ(Vector3D v, double r, Vector3D b) { return rotZ(v, r, b.x(), b.y(), b.z()); }

    /**
       Returns the vector rotated on x-axis around (bx, by, bz)
     */
    public Vector3D rotX(double rx, double bx, double by, double bz) { return rotX(this, rx, bx, by, bz); }

    /**
       Returns the vector rotated on y-axis around (bx, by, bz)
     */
    public Vector3D rotY(double rx, double bx, double by, double bz) { return rotY(this, rx, bx, by, bz); }

    /**
       Returns the vector rotated on z-axis around (bx, by, bz)
     */
    public Vector3D rotZ(double rx, double bx, double by, double bz) { return rotZ(this, rx, bx, by, bz); }

    /**
       Returns the vector rotated on x-axis around b
     */
    public Vector3D rotX(double r, Vector3D b) { return rotX(this, r, b.x(), b.y(), b.z()); }

    /**
       Returns the vector rotated on y-axis around b
     */
    public Vector3D rotY(double r, Vector3D b) { return rotY(this, r, b.x(), b.y(), b.z()); }

    /**
       Returns the vector rotated on z-axis around b
     */
    public Vector3D rotZ(double r, Vector3D b) { return rotZ(this, r, b.x(), b.y(), b.z()); }

    public static boolean epsilonEquals(Vector3D v, Vector3D w, double eps) {
	return distance(v, w) <= eps;
    }

    public static boolean epsilonEquals(Vector3D v, Vector3D w) {
	return distance(v, w) <= MathUtils.epsilon;
    }

    //

    /**
       Returns the length of this vector
     */
    public double length() {
	return Math.sqrt(x*x + y*y + z*z);
    }

    /**
       Returns the squared length of this vector
     */
    public double lengthSquared() {
	return x*x + y*y + z*z;
    }

    public boolean epsilonEquals(Vector3D v, double eps) {
	return distance(this,v) <= eps;
    }

    public boolean epsilonEquals(Vector3D v) {
	return distance(this,v) <= MathUtils.epsilon;
    }

    /**
       Returns this + v
     */
    public Vector3D plus(Vector3D v) {
	return add(this,v);
    }

    /**
       Returns this - v
     */
    public Vector3D minus(Vector3D v) {
	return sub(this,v);
    }

    public Vector3D normalized() {
	return normalize(this);
    }

    /**
       Returns the dot product of this and v
     */
    public double dot(Vector3D v) {
	return dotProduct(this,v);
    }

    /**
       Returns this vector scaled by d
     */
    public Vector3D dot(double d) {
	return scale(d,this);
    }

    /**
       Returns this vector scaled by d
     */
    public Vector3D times(double d) {
	return scale(d,this);
    }

    /**
       Returns the cross product of this and v
     */
    public Vector3D cross(Vector3D v) {
	return crossProduct(this,v);
    }

    //

    /**
       Returns hash code of this instance
     */
    public int hashCode() {
	/*
	SimpleHashFunc hash = new SimpleHashFunc();

	hash.add(x);
	hash.add(y);
	hash.add(z);

	return (int)hash.getHashValue();
	*/
	return (int)(Double.doubleToRawLongBits(x) ^ Double.doubleToRawLongBits(y) ^ Double.doubleToRawLongBits(z));
    }

    public static int hashCode(Vector3D v) {
	return (int)(Double.doubleToRawLongBits(v.x) ^ Double.doubleToRawLongBits(v.y) ^ Double.doubleToRawLongBits(v.z));
    }

    public int compareTo(Object o) {
	Vector3D v = (Vector3D) o;

	if (x > v.x) {
	    return 1;
	} else if (x < v.x) {
	    return -1;
	} else if (y > v.y) {
	    return 1;
	} else if (y < v.y) {
	    return -1;
	} else if (z > v.z) {
	    return 1;
	} else if (z < v.z) {
	    return -1;
	}

	return 0;
    }

    public boolean equals(Vector3D v) {
	return v.x == x && v.y == y && v.z == z;
    }

    public boolean equals(Object o) {
	return equals((Vector3D)o);
    }

    public static boolean equals(Vector3D v, Vector3D w) {
	return v.x == w.x && v.y == w.y && v.z == w.z;
    }

    /**
       Returns the centroid of collection c of vectors
     */
    public static Vector3D centroid(Iterable<? extends Vector3D> c) {
	int count = 0;
	Vector3D v = new Vector3D(0,0,0);

	for(Vector3D w : c) {
	    count++;
	    v = add(v, w);
	}

	return scale(1.0/count,v);
    }

    /**
       Convert this vector to a float array
     */
    public float[] toFloatArray() {
	float ret[] = new float[3];

	ret[0] = (float)x;
	ret[1] = (float)y;
	ret[2] = (float)z;

	return ret;
    }

    /**
       Convert this vector to a float array of length n
     */
    public float[] toFloatArray(int n) {
	float ret[] = new float[n];

	ret[0] = (float)x;
	ret[1] = (float)y;
	ret[2] = (float)z;

	return ret;
    }
}
